import logging
from abc import abstractmethod
from datetime import datetime

import apscheduler.events
import apscheduler.schedulers.base

from multiprocessing import RLock


class SchedulerEventsListener:

    @abstractmethod
    def _scheduler_event(self, event):
        """
        Event triggered whenever the scheduler status changes.

        Args:
            event (dict):
        """

    @abstractmethod
    def _job_event(self, event):
        """
        Event triggered when a job is added, modified, removed, when it's submitted for execution or when executing it
        finishes.

        Args:
            event (dict):
        """

    @abstractmethod
    def _jobstore_event(self, event):
        """
        Triggered any time an job store is added or removed from the scheduler.

        Args:
            event (dict):
        """

    @abstractmethod
    def _executor_event(self, event):
        """
        Triggered any time an executor is added or removed from the scheduler.

        Args:
            event (dict):
        """


class SchedulerWatcher:

    scheduler_states = {
        apscheduler.schedulers.base.STATE_RUNNING: 'running',
        apscheduler.schedulers.base.STATE_PAUSED: 'paused',
        apscheduler.schedulers.base.STATE_STOPPED: 'stopped'
    }

    apscheduler_events = {
        apscheduler.events.EVENT_SCHEDULER_STARTED: 'scheduler_started',
        apscheduler.events.EVENT_SCHEDULER_PAUSED: 'scheduler_paused',
        apscheduler.events.EVENT_SCHEDULER_SHUTDOWN: 'scheduler_shutdown',
        apscheduler.events.EVENT_SCHEDULER_RESUMED: 'scheduler_resumed',

        apscheduler.events.EVENT_EXECUTOR_ADDED: 'executor_added',
        apscheduler.events.EVENT_EXECUTOR_REMOVED: 'executor_removed',

        apscheduler.events.EVENT_JOBSTORE_ADDED: 'jobstore_added',
        apscheduler.events.EVENT_JOBSTORE_REMOVED: 'jobstore_removed',

        apscheduler.events.EVENT_ALL_JOBS_REMOVED: 'all_jobs_removed',
        apscheduler.events.EVENT_JOB_ADDED: 'job_added',
        apscheduler.events.EVENT_JOB_REMOVED: 'job_removed',
        apscheduler.events.EVENT_JOB_MODIFIED: 'job_modified',

        apscheduler.events.EVENT_JOB_EXECUTED: 'job_executed',
        apscheduler.events.EVENT_JOB_ERROR: 'job_error',
        apscheduler.events.EVENT_JOB_MISSED: 'job_missed',

        apscheduler.events.EVENT_JOB_SUBMITTED: 'job_submitted',
        apscheduler.events.EVENT_JOB_MAX_INSTANCES: 'job_max_instances'
    }

    def __init__(self, scheduler, max_events_per_job=100):
        """
        Inspects the scheduler, registers itself as a scheduler event listener and keeps track of all changes to the
        scheduler and its jobs.

        Args:
            scheduler (apscheduler.schedulers.base.BaseScheduler):
                A reference to the scheduler we'd like to watch.

            max_events_per_job (int):
                The maximum amount of events we'll keep in-memory for each job to send to the clients when they connect.
        """
        self.scheduler = scheduler
        self.listeners = []
        self.max_events_per_job = max_events_per_job

        self.jobstores = {}
        self.executors = {}
        self.scheduler_info = {}
        self.jobs = {}

        self.write_lock = RLock()

        with self.write_lock:
            # Append ourselves as listeners of the scheduler first.
            self.scheduler.add_listener(self._process_event, mask=apscheduler.events.EVENT_ALL)
            # Inspect scheduler to init our attributes.
            self._inspect_scheduler()

    def add_listener(self, listener: SchedulerEventsListener):
        if listener not in self.listeners:
            self.listeners.append(listener)

    def _inspect_scheduler(self):
        if hasattr(self.scheduler, '_jobstores') and isinstance(self.scheduler._jobstores, dict):
            self.jobstores = {
                alias: self._repr_jobstore(store) for alias, store in self.scheduler._jobstores.items()
            }

        if hasattr(self.scheduler, '_executors') and isinstance(self.scheduler._executors, dict):
            self.executors = {
                alias: self._repr_executor(executor) for alias, executor in self.scheduler._executors.items()
            }

        self.scheduler_info.update({
            'class': type(self.scheduler).__name__,
            'state': self.scheduler_states[self.scheduler.state],
            'timezone': str(self.scheduler.timezone),
            'jobstore_retry_interval': self.scheduler.jobstore_retry_interval
        })

        if hasattr(self.scheduler, '_job_defaults'):
            self.scheduler_info.update(self.scheduler._job_defaults)

        init_ts = self._repr_ts(datetime.now(tz=self.scheduler.timezone))

        if len(self.jobstores) > 0:
            for jobstore_alias in self.jobstores.keys():
                self._jobstore_added(jobstore_alias, init_ts)
        else:
            for job in self.scheduler.get_jobs():
                if job.id not in self.jobs:
                    self._job_added(job.id, job.jobstore, init_ts, job)

    def _process_event(self, event):
        """
        Scheduler events listener. Dispatches events to the corresponding method.
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        if event.code in self.apscheduler_events:
            event_ts = self._repr_ts(datetime.now(tz=self.scheduler.timezone))
            event_name = self.apscheduler_events[event.code]
            self.__getattribute__(event_name)(event, event_name, event_ts)

    def scheduler_started(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            self.scheduler_info['state'] = self.scheduler_states[self.scheduler.state]

        self.notify_scheduler_event(event_name, event_ts)

    def scheduler_shutdown(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            self.scheduler_info['state'] = self.scheduler_states[self.scheduler.state]

        self.notify_scheduler_event(event_name, event_ts)

    def scheduler_paused(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            self.scheduler_info['state'] = self.scheduler_states[self.scheduler.state]

        self.notify_scheduler_event(event_name, event_ts)

    def scheduler_resumed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            self.scheduler_info['state'] = self.scheduler_states[self.scheduler.state]

        self.notify_scheduler_event(event_name, event_ts)

    def executor_added(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        executor = None
        if hasattr(self.scheduler, '_executors') and isinstance(self.scheduler._executors, dict):
            try:
                executor = self.scheduler._executors[event.alias]
            except KeyError:
                logging.getLogger('apschedulerui').warning('Failed to locate executor "%s" in scheduler' % event.alias)

        with self.write_lock:
            self.executors[event.alias] = self._repr_executor(executor) if executor else None

        self.notify_executor_event(event_name, event_ts, event.alias)

    def executor_removed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            if event.alias in self.executors:
                del self.executors[event.alias]

        self.notify_executor_event(event_name, event_ts, event.alias)

    def jobstore_added(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        self._jobstore_added(event.alias, event_ts)

        self.notify_jobstore_event(event_name, event_ts, event.alias)

    def jobstore_removed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            for job_id in self.jobs.keys():
                if self.jobs[job_id]['properties']['jobstore'] == event.alias:
                    self._job_removed(job_id, event_ts)

            if event.alias in self.jobstores:
                del self.jobstores[event.alias]

        self.notify_jobstore_event(event_name, event_ts, event.alias)

    def all_jobs_removed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.SchedulerEvent):
        """
        with self.write_lock:
            for job_id in self.jobs.keys():
                self._job_removed(job_id, removal_ts=event_ts)

    def job_added(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobEvent):
        """
        self._job_added(event.job_id, event.jobstore, event_ts)

    def job_removed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobEvent):
        """
        self._job_removed(event.job_id, removal_ts=event_ts)

    def job_modified(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobEvent):
        """
        self._job_modified(event.job_id, event.jobstore, event_ts)

    def job_executed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobExecutionEvent):
        """
        self._job_execution_event(event.job_id, event.jobstore, event_name, event_ts,
                                  retval=event.retval,
                                  scheduled_run_time=self._repr_ts(event.scheduled_run_time))

    def job_error(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobExecutionEvent):
        """
        self._job_execution_event(event.job_id, event.jobstore, event_name, event_ts,
                                  retval=event.retval,
                                  exception=str(event.exception),
                                  traceback=str(event.traceback),
                                  scheduled_run_time=self._repr_ts(event.scheduled_run_time))

    def job_missed(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobExecutionEvent):
        """
        self._job_execution_event(event.job_id, event.jobstore, event_name, event_ts,
                                  scheduled_run_time=self._repr_ts(event.scheduled_run_time))

    def job_submitted(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobSubmissionEvent):
        """
        self._job_execution_event(event.job_id, event.jobstore, event_name, event_ts,
                                  scheduled_run_time=self._repr_ts(event.scheduled_run_times[0]))

    def job_max_instances(self, event, event_name, event_ts):
        """
        Args:
            event (apscheduler.events.JobSubmissionEvent):
        """
        self._job_execution_event(event.job_id, event.jobstore, event_name, event_ts,
                                  scheduled_run_time=self._repr_ts(event.scheduled_run_times[0]))

    def scheduler_summary(self):
        return {
            'executors': {name: str(executor) for name, executor in self.executors.items()},
            'jobstores': {name: str(jobstore) for name, jobstore in self.jobstores.items()},
            'scheduler': self.scheduler_info,
            'jobs': self.jobs
        }

    def notify_scheduler_event(self, event_name, event_ts):
        for listener in self.listeners:
            listener._scheduler_event({
                'event_name': event_name,
                'event_ts': event_ts
            })

    def notify_job_event(self, event):
        event['next_run_times'] = []

        job = self.scheduler.get_job(job_id=event['job_id'])

        if job:
            next_run_time = job.next_run_time
            while next_run_time and len(event['next_run_times']) < 11:
                event['next_run_times'].append(self._repr_ts(next_run_time))
                next_run_time = job.trigger.get_next_fire_time(next_run_time, next_run_time)

        for listener in self.listeners:
            listener._job_event(event)

    def notify_executor_event(self, event_name, event_ts, executor_name):
        for listener in self.listeners:
            listener._executor_event({
                'event_name': event_name,
                'event_ts': event_ts,
                'executor_name': executor_name
            })

    def notify_jobstore_event(self, event_name, event_ts, jobstore_name):
        for listener in self.listeners:
            listener._jobstore_event({
                'event_name': event_name,
                'event_ts': event_ts,
                'jobstore_name': jobstore_name
            })

    def _job_added(self, job_id, jobstore, added_ts, job=None):
        with self.write_lock:
            if job_id in self.jobs:
                return

            self.jobs[job_id] = {
                'added_time': added_ts,
                'modified_time': added_ts,
                'removed_time': None,
                'properties': self._repr_job(
                    self.scheduler.get_job(job_id, jobstore) if job is None else job,
                    jobstore=jobstore
                ),
                'events': []
            }

            event = {
                'job_id': job_id,
                'event_name': 'job_added',
                'event_ts': added_ts
            }

            event.update(self.jobs[job_id])

            self._append_job_event(event)

        self.notify_job_event(event)

    def _job_modified(self, job_id, jobstore, event_ts):
        with self.write_lock:
            self.jobs[job_id]['properties'] = self._repr_job(
                self.scheduler.get_job(job_id, jobstore),
                jobstore=jobstore
            )
            self.jobs[job_id]['modified_time'] = event_ts

            event = {
                'job_id': job_id,
                'event_name': 'job_modified',
                'event_ts': event_ts,
                'properties': self.jobs[job_id]['properties']
            }

            self._append_job_event(event)

        self.notify_job_event(event)

    def _job_removed(self, job_id, removal_ts):
        with self.write_lock:
            self.jobs[job_id]['removed_time'] = removal_ts

            event = {
                'job_id': job_id,
                'event_name': 'job_removed',
                'event_ts': removal_ts
            }

            self._append_job_event(event)

        self.notify_job_event(event)

    def _job_execution_event(self, job_id, jobstore, event_name, event_ts, **kwargs):
        with self.write_lock:
            if job_id not in self.jobs:
                self._job_added(job_id, jobstore, event_ts)

            event = {
                'job_id': job_id,
                'event_name': event_name,
                'event_ts': event_ts,
                'next_run_times': []
            }
            event.update(kwargs)

            self._append_job_event(event)

        self.notify_job_event(event)

    def _jobstore_added(self, jobstore, event_ts):
        with self.write_lock:
            for job in self.scheduler.get_jobs(jobstore=jobstore):
                self._job_added(job.id, jobstore, event_ts, job)

            if hasattr(self.scheduler, '_jobstores') and isinstance(self.scheduler._jobstores, dict):
                self.jobstores[jobstore] = self._repr_jobstore(self.scheduler._jobstores[jobstore])

    def _repr_job(self, job, jobstore=None):
        next_run_time = self._repr_ts(getattr(job, 'next_run_time', None))
        return {
            'id': job.id,
            'name': job.name,
            'trigger': self._repr_trigger(job.trigger),
            'jobstore': jobstore,
            'executor': job.executor,
            'func': str(job.func),
            'func_ref': job.func_ref,
            'args': str(job.args),
            'kwargs': str(job.kwargs),
            'pending': job.pending,
            'coalesce': getattr(job, 'coalesce', None),
            'next_run_time': [next_run_time] if next_run_time else None,
            'misfire_grace_time': getattr(job, 'misfire_grace_time', None),
            'max_instances': getattr(job, 'max_instances', None)
        }

    def _append_job_event(self, e):
        if 'events' in e:
            del e['events']  # Make sure we won't be appending to the event list a reference to itself!

        job_id = e['job_id']

        self.jobs[job_id]['events'].append(e)

        if len(self.jobs[job_id]['events']) > self.max_events_per_job:
            # Limit job event log size.
            self.jobs[job_id]['events'] = self.jobs[job_id]['events'][-self.max_events_per_job:]

    def _repr_ts(self, ts):
        """
        Args:
            ts (datetime):

        Returns:
            str: The timestamp string representation.
        """
        if ts:
            return ts.strftime('%Y-%m-%d %H:%M:%S.%f')
        return None

    def _repr_trigger(self, trigger):
        return str(trigger)

    def _repr_jobstore(self, jobstore):
        return str(jobstore)

    def _repr_executor(self, jobstore):
        return str(jobstore)



from setuptools import setup, find_packages
from setuptools.command.sdist import sdist

VERSION = '0.0.1'

with open('README.md', 'r') as f:
    long_description = f.read()

dependencies = [
    'apscheduler >= 3.0.0',
    'flask >= 1.0.0',
    'flask_socketio >= 4.0.0'
]


class SdistWithWebpack(sdist):
    """
    We need to extend the sdist command to run webpack in production mode before building the package for distribution.
    """

    def run(self):
        import os
        import subprocess

        if 'CI' not in os.environ:
            if not os.path.exists('frontend/node_build_env'):
                subprocess.check_call(['nodeenv', 'node_build_env'], cwd='frontend')

            subprocess.check_call(
                ['./node_build_env/bin/activate && npm install && npx webpack -p'],
                shell=True,
                cwd='frontend'
            )

        sdist.run(self)


setup(
    name='apschedulerui',
    version=VERSION,
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    author='Federico Schmidt',
    author_email='schmidt.fdr@gmail.com',
    url='https://github.com/schmidtfederico/apscheduler-ui',
    packages=find_packages(exclude=['tests']),
    install_requires=dependencies,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        "Programming Language :: Python :: 3.5",
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Operating System :: OS Independent',
        'Topic :: Internet'
    ],
    python_requires='>=3.5, <4',
    extras_require={
        'testing': ['requests'],
        'testing:python_version == "3.5"': ['mock']
    },
    cmdclass={'sdist': SdistWithWebpack},
    include_package_data=True
)



import logging
import threading
from datetime import datetime

import flask
import flask_socketio
from apscheduler.jobstores.base import JobLookupError
from apscheduler.triggers.cron import CronTrigger
from apscheduler.triggers.interval import IntervalTrigger
from flask import Response

from apschedulerui.watcher import SchedulerWatcher, SchedulerEventsListener


class SchedulerUI(SchedulerEventsListener):
    """
    A web server that monitors your scheduler and serves a web application to visualize events.

    By the default the server web application served is a view-only UI, but by enabling capabilities it may be allowed
    to control the scheduler and its jobs.

    Args:
        scheduler (apscheduler.schedulers.base.BaseScheduler):
            The scheduler to monitor.

        capabilities (dict):
            (Optional)
            A dictionary of the capabilities to enable in the server and client. By default the UI is view-only.

            Supported capabilities:
                * Pause/Resume Scheduler: set `pause_scheduler` to :data:`True`.
                * Pause/Resume Jobs: set `pause_job` to :data:`True`.
                * Remove Jobs: set `remove_job` to :data:True.

        operation_timeout (float):
            (Optional) The amount of seconds to wait for the serializing lock when performing actions on the
            scheduler or on its jobs from the UI.

    Basic Usage:
      >>> from apscheduler.schedulers.background import BackgroundScheduler
      >>> from apschedulerui.web import SchedulerUI
      >>> scheduler = BackgroundScheduler()
      >>> ui = SchedulerUI(scheduler)
      >>> ui.start()  # Server available at localhost:5000.

    Configuring capabilities:
      >>> ui = SchedulerUI(scheduler, capabilities={'pause_scheduler': True})  # All omitted capabilities are False.
      >>> ui = SchedulerUI(scheduler, capabilities={'pause_job': True, 'remove_job': True})

    """

    def __init__(self, scheduler, capabilities=None, operation_timeout=1):
        self.scheduler = scheduler
        self.capabilities = {
            'pause_job': False,
            'remove_job': False,
            'pause_scheduler': False,
            'stop_scheduler': False,
            'run_job': False,
        }

        if not (isinstance(operation_timeout, int) or isinstance(operation_timeout, float)):
            raise TypeError('operation_timeout should be either an int or a float')

        if operation_timeout <= 0:
            raise ValueError('operation_timeout should be a positive number')

        self.operation_timeout = operation_timeout

        if capabilities is not None:
            if isinstance(capabilities, dict):
                self.capabilities.update(capabilities)
            else:
                raise TypeError('capabilities should be a dict of str -> bool pairs')

        self._scheduler_listener = SchedulerWatcher(scheduler)

        self._web_server = flask.Flask(__name__)
        self._socket_io = None

        try:
            # TODO: see if we can support eventlet in the future.
            self._socket_io = flask_socketio.SocketIO(self._web_server, async_mode='gevent')
        except ValueError:
            self._socket_io = flask_socketio.SocketIO(self._web_server, async_mode='threading')

        self._init_endpoints()

        self._web_server_thread = None
        self._scheduler_lock = threading.Lock()

    def start(self, host='0.0.0.0', port=5000, daemon=True):
        """
        Starts listening for events from the scheduler and starts the web server that serves the UI in a new thread.

        Args:
            host (str):
                (Optional) The address to bind the web server to. Default `0.0.0.0`.
            port (int):
                (Optional) The port to which the web server will bind. Defaults to :data:`5000`.
            daemon (bool):
                (Optional) If :data:`True` (default) starts the server as daemon.

        """
        self._scheduler_listener.add_listener(self)
        self._web_server_thread = threading.Thread(target=self._start, name='apscheduler-ui', args=(host, port))
        self._web_server_thread.daemon = daemon
        self._web_server_thread.start()

    def _init_endpoints(self):
        if self.capabilities.get('pause_scheduler', False):
            self._web_server.add_url_rule(
                '/api/scheduler/pause', 'pause_scheduler', self._pause_scheduler, methods=['POST']
            )
            self._web_server.add_url_rule(
                '/api/scheduler/resume', 'resume_scheduler', self._resume_scheduler, methods=['POST']
            )

        if self.capabilities.get('stop_scheduler', False):
            self._web_server.add_url_rule(
                '/api/scheduler/stop', 'stop_scheduler', self._stop_scheduler, methods=['POST']
            )
            self._web_server.add_url_rule(
                '/api/scheduler/start', 'start_scheduler', self._start_scheduler, methods=['POST']
            )

        if self.capabilities.get('remove_job', False):
            self._web_server.add_url_rule('/api/job/<job_id>/remove', 'remove_job', self._remove_job, methods=['POST'])

        if self.capabilities.get('pause_job', False):
            self._web_server.add_url_rule('/api/job/<job_id>/pause', 'pause_job', self._pause_job, methods=['POST'])
            self._web_server.add_url_rule('/api/job/<job_id>/resume', 'resume_job', self._resume_job, methods=['POST'])

        if self.capabilities.get('run_job', False):
            self._web_server.add_url_rule('/api/job/<job_id>/run_now', 'run_job', self._run_job, methods=['POST'])

        self._web_server.add_url_rule('/', 'index', self._index, defaults={'path': ''})
        self._web_server.add_url_rule('/<path:path>', 'index', self._index)

        self._socket_io.on_event('connected', self._client_connected)

    def _index(self, path):
        return self._web_server.send_static_file('index.html')

    def _exec_scheduler_command(self, func, *args, **kwargs):
        if self._scheduler_lock.acquire(timeout=self.operation_timeout):
            try:
                func(*args, **kwargs)
                return 'ok'
            except JobLookupError:
                flask.abort(404, description="Job not found")
            finally:
                self._scheduler_lock.release()
        else:
            flask.abort(408, description="Failed to acquire scheduler lock to perform operation")

    def _pause_scheduler(self):
        return self._exec_scheduler_command(self.scheduler.pause)

    def _resume_scheduler(self):
        return self._exec_scheduler_command(self.scheduler.resume)

    def _stop_scheduler(self):
        return self._exec_scheduler_command(self.scheduler.shutdown, wait=False)

    def _start_scheduler(self):
        return self._exec_scheduler_command(self.scheduler.start)

    def _pause_job(self, job_id):
        return self._exec_scheduler_command(self.scheduler.pause_job, job_id)

    def _resume_job(self, job_id):
        return self._exec_scheduler_command(self.scheduler.resume_job, job_id)

    def _run_job(self, job_id, next_run_time=None):
        logging.getLogger('apschedulerui').info('Running job %s' % job_id)
        if not job_id:
            return Response(status=404)

        if not next_run_time:
            next_run_time = datetime.now()

        def _run_job_impl():
            job = self.scheduler.get_job(job_id)

            if not job:
                raise JobLookupError(job_id)

            # If a job is periodic (has an interval trigger) it should be triggered by modifying the trigger it already
            # has. Otherwise, it can be rescheduled to be ran now.
            if isinstance(job.trigger, IntervalTrigger) or isinstance(job.trigger, CronTrigger):
                self.scheduler.modify_job(job_id, next_run_time=next_run_time)
            else:
                job.reschedule(trigger='date', run_date=next_run_time)

        return self._exec_scheduler_command(_run_job_impl)

    def _remove_job(self, job_id):
        return self._exec_scheduler_command(self.scheduler.remove_job, job_id)

    def _client_connected(self):
        logging.getLogger('apschedulerui').debug('Client connected')
        flask_socketio.emit('init_jobs', self._scheduler_listener.scheduler_summary())
        flask_socketio.emit('init_capabilities', self.capabilities)

    def _job_event(self, event):
        self._socket_io.emit('job_event', event)

    def _scheduler_event(self, event):
        self._socket_io.emit('scheduler_event', event)

    def _jobstore_event(self, event):
        self._socket_io.emit('jobstore_event', event)

    def _executor_event(self, event):
        self._socket_io.emit('executor_event', event)

    def _start(self, host, port):
        self._socket_io.run(self._web_server, host=host, port=port)
